#############################################################################
# puo_plugin.tcl Copyright 2005, 2006, 2007, 2008, 2009 r0lZ
# PgcEdit plugin to import, export and clear prohibited user operations.
#
# This plugin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This plugin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################
# Usage: copy this file in the <PgcEdit's install dir>/plugins directory.
# A "Prohibited User Operations" menu will be added in the Plugins menu.
# You need PgcEdit 0.6.0 or more to use this plugin.
# It is better to use v 0.6.1 or more.
#############################################################################
# Version History:
#
# v 2.0 (August 26 2009)
# - Added "Edit VOB PUOs of current domain" and "Edit VOB PUOs of current PGC"
#   with a GUI to selectively set or clear any PUO of any cell of the current
#   domain or PGC.
# - Removed Selectively Clear VOB PUOs, less powerful than the new functions.
# - The plugin menu has been reorganised.
# v 1.2 (September 7, 2005)
# - First public release
# v 1.3 (November 3, 2005)
# - It is now possible to selectively keep some PUOs when removing them from
#   the VOB files.
# v 1.4 (October 7, 2006)
# - Bug fixed.  A wrong error message was displayed in some rare cases, when
#   the end of a VOB file is padded with null packs.
#############################################################################



# Add Plugins -> PUO submenu
if {! [winfo exists .mainmenu.plugins.iepuos]} {
	.mainmenu.plugins add cascade -label "Prohibited User Operations" -menu .mainmenu.plugins.iepuos
	menu .mainmenu.plugins.iepuos
	.mainmenu.plugins.iepuos add command -label "Export all IFO PUOs" -command {::plugins::iepuos::export_ifo_puos}
	.mainmenu.plugins.iepuos add command -label "Import all IFO PUOs" -command {::plugins::iepuos::import_ifo_puos}
	.mainmenu.plugins.iepuos add command -label "Clear all IFO PUOs"  -command {::plugins::iepuos::clear_ifo_puos}
	.mainmenu.plugins.iepuos add command -label "Edit IFO PUOs of all PGCs of current domain"  -command {::plugins::iepuos::set_domain_puos_global}
	.mainmenu.plugins.iepuos add separator
	.mainmenu.plugins.iepuos add command -label "Edit VOB PUOs of current domain (by V/CID)"  -command {::plugins::iepuos::set_domain_puos_by_vcid "domain"}
	.mainmenu.plugins.iepuos add command -label "Edit VOB PUOs of current PGC (by V/CID)"  -command {::plugins::iepuos::set_domain_puos_by_vcid "pgc"}
	.mainmenu.plugins.iepuos add separator
	.mainmenu.plugins.iepuos add command -label "About" -command {
			tk_messageBox  -title "Prohibited User Operations plugin" -icon info \
					-message "Prohibited User Operations plugin for PgcEdit 9.3 or more.\nVersion $::plugins::iepuos::version\nAuthor: r0lZ\nFreeware!"
		}
}

namespace eval ::plugins {}
namespace eval ::plugins::iepuos {

set version "2.0"

# Export IFO PUOs
proc export_ifo_puos {} {
	if {![info exists ::dvddir]} {bell; return}
	set title "PUO plugin - Export IFO PUOs          "

	set filename [tk_getSaveFile -filetypes {{"Exported PUOs" {*.puo}}} \
		-parent . -initialdir $::config(import_initialdir) \
		-initialfile "PUOs.puo" -title $title]
	if {$filename == ""} {
		return
	}
	if {[string toupper [file extension $filename]] != ".PUO"} {
		set filename "${filename}.puo"
	}
	set ::config(import_initialdir) [file dirname $filename]

	set numpgcs 1

	set txt "# Exported Prohibited User Operations\n"
	set txt "${txt}# DVD: \"$::dvddir\"\n"
	set txt "${txt}# DVD-TEXT General Name: \"[::utils::get_General_Name]\"\n"
	set txt "${txt}# Provider ID: \"[::utils::get_provider_ID]\"\n"
	set txt "${txt}# Number of VTS: $::pgcs(numvts)\n\n"

	set txt "${txt}set plugin_version $::plugins::iepuos::version\n\n"

	set txt "${txt}# PUOs in VMG_TT_SRPT (Table of Titles)\n"
	set tt_srpt_puos false
	set numtitles [get_int ::raw_sectors(0,VMG_TT_SRPT) 0 2]
	set txt "${txt}set puos(numtitles) $numtitles\n"
	for {set i 1} {$i <= $numtitles} {incr i} {
		set puos [expr [get_int ::raw_sectors(0,VMG_TT_SRPT) [expr {8+(($i-1)*12)}] 1] & 3]
		set txt "${txt}set puos(title,$i) $puos\n"
		if {$puos != 0} {set tt_srpt_puos true}
	}
	set txt "${txt}set puos(title,exists) $tt_srpt_puos\n\n\n"

	set puos [get_int ::pgcs(0,0,header) 8 4]
	set txt "${txt}# [::main::pgc_label $::config(fixed_width_font) pgc 0 0]\n"
	set txt "${txt}set puos(pgc,0,0) $puos\t;# 0x[format %08X $puos]\n"

	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		set txt "${txt}\n"
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set puos [get_int ::menus($vts,$lu,$pgc,header) 8 4]
				set txt "${txt}# [::main::pgc_label $::config(fixed_width_font) menu $vts $lu $pgc]\n"
				set txt "${txt}set puos(menu,$vts,$lu,$pgc) $puos\t;# 0x[format %08X $puos]\n"
				incr numpgcs
			}
		}
		set txt "${txt}\n"
		if {$vts > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set puos [get_int ::pgcs($vts,$pgc,header) 8 4]
				set txt "${txt}# [::main::pgc_label $::config(fixed_width_font) pgc $vts $pgc]\n"
				set txt "${txt}set puos(pgc,$vts,$pgc) $puos\t;# 0x[format %08X $puos]\n"
				incr numpgcs
			}
		}
	}
	set txt "${txt}\n# Number of PGCs in DVD: $numpgcs\n"

	if {! [saveFile $filename auto $txt]} {
		show_msg "Prohibited User Operations successfully exported from $numpgcs PGCs and $numtitles Titles in VMG_TT_SRPT."
	} else {
		show_msg "ERROR exporting Prohibited User Operations from $numpgcs PGCs!"
	}
}

# Import IFO PUOs
proc import_ifo_puos {} {
	if {![info exists ::dvddir]} {bell; return}
	set title "PUO plugin - Import IFO PUOs          "

	set filename [tk_getOpenFile -filetypes {{"Exported PUOs" {*.puo}}} \
		-parent . -initialdir $::config(import_initialdir) \
		-initialfile "PUOs.puo" -title $title]
	if {$filename == ""} {
		return
	}
	set ::config(import_initialdir) [file dirname $filename]

	if {[catch {source $filename} errstr]} {
		show_msg "ERROR importing PUOs!"
		tk_messageBox -title $title -icon error -type ok \
			-message "Error sourcing \"$filename\":\n$errstr"
		return
	}

	# PUOs in VMG_TT_SRPT
	set numtitles [get_int ::raw_sectors(0,VMG_TT_SRPT) 0 2]
	set puotitles 0
	for {set i 1} {$i <= $numtitles} {incr i} {
		if {[info exists puos(title,$i)]} {
			set tidx [expr {8+(($i-1)*12)}]
			set titletype [expr ([get_int ::raw_sectors(0,VMG_TT_SRPT) $tidx 1] & 0xFC) | $puos(title,$i)]
			set_int ::raw_sectors(0,VMG_TT_SRPT) $tidx $titletype 1
			incr puotitles
		}
	}

	# PUOs in PGCs
	set numpgcs 1
	set imported 0
	if {[info exists puos(pgc,0,0)]} {
		set_int ::pgcs(0,0,header) 8 $puos(pgc,0,0) 4
		incr imported
	}
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				incr numpgcs
				if {[info exists puos(menu,$vts,$lu,$pgc)]} {
					set_int ::menus($vts,$lu,$pgc,header) 8 $puos(menu,$vts,$lu,$pgc) 4
					incr imported
				}
			}
		}
		if {$vts > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				incr numpgcs
				if {[info exists puos(pgc,$vts,$pgc)]} {
					set_int ::pgcs($vts,$pgc,header) 8 $puos(pgc,$vts,$pgc) 4
					incr imported
				}
			}
		}
	}
	show_msg "PUOs successfully imported in $imported / $numpgcs PGCs and $puotitles / $numtitles Titles."
	::main::dvd_modified true false
}

# Clear IFO PUOs
proc clear_ifo_puos {} {
	if {![info exists ::dvddir]} {bell; return}
	set title "PUO plugin - Clear IFO PUOs          "

	if {! $::config(expert_mode) && [tk_messageBox -title $title -icon question -type okcancel \
			-default ok -message "Clear all Prohibited User Operations in IFOs?\nAre you sure?"] == "cancel"} return

	# PUOs in VMG_TT_SRPT
	set numtitles [get_int ::raw_sectors(0,VMG_TT_SRPT) 0 2]
	for {set i 1} {$i <= $numtitles} {incr i} {
		set tidx [expr {8+(($i-1)*12)}]
		set titletype [expr [get_int ::raw_sectors(0,VMG_TT_SRPT) $tidx 1] & 0xFC]
		set_int ::raw_sectors(0,VMG_TT_SRPT) $tidx $titletype 1
	}

	set numpgcs 1
	set_int ::pgcs(0,0,header) 8 0 4
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				incr numpgcs
				set_int ::menus($vts,$lu,$pgc,header) 8 0 4
			}
		}
		if {$vts > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				incr numpgcs
				set_int ::pgcs($vts,$pgc,header) 8 0 4
			}
		}
	}
	show_msg "Prohibited User Operations successfully cleared in $numpgcs PGCs and $numtitles Titles."
	::main::dvd_modified true false
}

# Set domain PUOs
proc set_domain_puos_global {} {
	if {![info exists ::dvddir]} {bell; return}
	set type [lindex $::editor_cmds_from 0]
	set vts [lindex [split [lindex $::editor_cmds_from 1] ,] 0]
	#if {$type == "pgc" && $vts == 0} {bell; return}

	set ::pgced::puos [expr 0xFFFFFFFF]
	if {$type == "pgc"} {
		if {$vts == 0} {
			set ::pgced::puos [get_int ::pgcs(0,0,header) 8 4]
		} else {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set ::pgced::puos [expr $::pgced::puos & [get_int ::pgcs($vts,$pgc,header) 8 4]]
			}
		}
	} else {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set ::pgced::puos [expr $::pgced::puos & [get_int ::menus($vts,$lu,$pgc,header) 8 4]]
			}
		}
	}
	if {[catch {set puos [::pgced::puos_editor "PUO plugin\nSet PUOs for all PGCs in current domain" . false]}]} {
		set ::pgced::puos -1
		set puos [::pgced::puos_editor "PUO plugin\nSet PUOs for all PGCs in current domain" .]
	}
	if {$puos == -1} return

	set numpgcs 0
	if {$type == "pgc"} {
		if {$vts == 0} {
			set_int ::pgcs(0,0,header) 8 $puos 4
			incr numpgcs
		} else {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set_int ::pgcs($vts,$pgc,header) 8 $puos 4
				incr numpgcs
			}
		}
	} else {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set_int ::menus($vts,$lu,$pgc,header) 8 $puos 4
				incr numpgcs
			}
		}
	}
	show_msg "Prohibited User Operations successfully set in $numpgcs PGCs."
	::main::dvd_modified true false
}


################## Set Domain PUOs by VOB/Cell ID (with GUI) ##################
###################### (Requires PgcEdit 9.2 or greater) ######################

# scope can be "domain" or "pgc"
proc set_domain_puos_by_vcid {scope} {
	if {! [verify_dvd_loaded]} {return}
	if {![test_version 9.3 "PUO Plugin (VOB functions)"]} return

	namespace eval ::plugins::iepuos::vars {}

	set type [lindex $::editor_cmds_from 0]
	set vts [lindex [split [lindex $::editor_cmds_from 1] ,] 0]
	set all_vcids {}
	set ::plugins::iepuos::vars::angle(all) false
	if {$type == "pgc"} {
		if {$vts == 0} {
			bell; return
		}
		if {$scope == "domain"} {
			set lab "VTST $vts"
			set first_pgc 1
			set last_pgc  $::pgcs($vts,numpgcs)
		} else {
			set pgc [lindex [split [lindex $::editor_cmds_from 1] ,] 1]
			set lab "VTST $vts, PGC $pgc"
			set first_pgc $pgc
			set last_pgc  $pgc
		}
		for {set pgc $first_pgc} {$pgc <= $last_pgc} {incr pgc} {
			set cellpbtable  $::pgcs($vts,$pgc,cellpbtable)
			set cellpostable $::pgcs($vts,$pgc,cellpostable)
			for {set cell 0} {$cell < $::pgcs($vts,$pgc,numcells)} {incr cell} {
				set vid [get_int cellpostable [expr $cell*4] 2]
				set cid [get_int cellpostable [expr ($cell*4)+3] 1]
				set start [get_int cellpbtable [expr ($cell*24)+8] 4]
				if {[info exists ::buttons::butsinfo($type,$vts,$vid,$cid,navpack)]} {
					set ::plugins::iepuos::vars::lbas($vid,$cid) $::buttons::butsinfo($type,$vts,$vid,$cid,navpack)
				} else {
					set ::plugins::iepuos::vars::lbas($vid,$cid) $start
				}
				set ::plugins::iepuos::vars::start_end($vid,$cid) [list $start [get_int cellpbtable [expr ($cell*24)+20] 4]]
				if {[expr [get_int cellpbtable [expr $cell*24] 1] & 0xC0] > 0} {
					set ::plugins::iepuos::vars::angle($vid,$cid) true
					set ::plugins::iepuos::vars::angle(all) true
				} else {
					set ::plugins::iepuos::vars::angle($vid,$cid) false
				}
				lappend all_vcids [list $vid $cid]
			}
		}
	} else {
		if {$vts == 0} {
			set lab "VMGM"
		} else {
			set lab "VTSM $vts"
		}
		if {$scope == "domain"} {
			set first_lu  1
			set last_lu   $::menus($vts,numlus)
			set first_pgc 1
			# last_pgc is set later
		} else {
			set lu [lindex [split [lindex $::editor_cmds_from 1] ,] 1]
			set pgc [lindex [split [lindex $::editor_cmds_from 1] ,] 2]
			set lab "$lab, LU $lu, PGC $pgc"
			set first_lu  $lu
			set last_lu   $lu
			set first_pgc $pgc
			set last_pgc  $pgc
		}
		for {set lu $first_lu} {$lu <= $last_lu} {incr lu} {
			if {$scope == "domain"} {
				set last_pgc  $::menus($vts,$lu,numpgcs)
			}
			for {set pgc $first_pgc} {$pgc <= $last_pgc} {incr pgc} {
				set cellpbtable  $::menus($vts,$lu,$pgc,cellpbtable)
				set cellpostable $::menus($vts,$lu,$pgc,cellpostable)
				for {set cell 0} {$cell < $::menus($vts,$lu,$pgc,numcells)} {incr cell} {
					set vid [get_int cellpostable [expr $cell*4] 2]
					set cid [get_int cellpostable [expr ($cell*4)+3] 1]
					set start [get_int cellpbtable [expr ($cell*24)+8] 4]
					if {[info exists ::buttons::butsinfo($type,$vts,$vid,$cid,navpack)]} {
						set ::plugins::iepuos::vars::lbas($vid,$cid) $::buttons::butsinfo($type,$vts,$vid,$cid,navpack)
					} else {
						set ::plugins::iepuos::vars::lbas($vid,$cid) $start
					}
					set ::plugins::iepuos::vars::start_end($vid,$cid) [list $start [get_int cellpbtable [expr ($cell*24)+20] 4]]
					set ::plugins::iepuos::vars::angle($vid,$cid) false
					lappend all_vcids [list $vid $cid]
				}
			}
		}
	}
	if {[llength $all_vcids] == 0} {
		bell
		return
	}
	set all_vcids [lsort -dictionary $all_vcids]
	set vcids_list {}
	set last "-"
	foreach i $all_vcids {
		if {$i != $last} {
			lappend vcids_list $i
			set last $i
		}
	}

	set w .setvcidpuos
	toplevel $w -class Dialog ;# -menu .setvcidpuosmenu
	wm withdraw $w
	catch {grab set $w}
	wm protocol $w WM_DELETE_WINDOW {set ::plugins::iepuos::vars::done 0}
	if {$::config(tk) != "aqua"} {wm transient $w .}
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	bind $w <Escape> {set ::plugins::iepuos::vars::done 0}
	#bind $w <Return> {set ::plugins::iepuos::vars::done 1}

	### Main table
	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 2 -anchor w -fill x

	frame $w.f1.ft
	pack  $w.f1.ft -fill both

	frame  $w.f1.ft.puo -bd 3 -relief flat
	pack   $w.f1.ft.puo -side left -expand 0 -fill y -anchor w
	label  $w.f1.ft.puo.lab -text "VOB PUOs of $lab" -pady 0
	grid   $w.f1.ft.puo.lab -column 0 -row 1

	if {$::config(use_preview)} {
		button $w.f1.ft.puo.pre -image ::preview::bitmapplay -width 23 -height 15 \
				-command ::preview::preview_current_domain
		grid   $w.f1.ft.puo.pre -column 0 -row 2
	}

	checkbutton $w.f1.ft.puo.sel -text "Select/Deselect All Cells ->" \
		   -variable ::plugins::iepuos::vars::all_vcids -command ::plugins::iepuos::select_all_vcids \
		   -bd $::config(min_bd)
	set ::plugins::iepuos::vars::all_vcids 1
	grid   $w.f1.ft.puo.sel  -column 0 -row 3 -sticky e
	frame  $w.f1.ft.puo.sep -width 10 -height 2 -bd 2 -relief sunken
	grid   $w.f1.ft.puo.sep  -column 0 -row 4 -sticky ew
	set puon 0
	foreach p $::pgced::puos_labels {
		checkbutton $w.f1.ft.puo.puo$puon -text $p -anchor w -pady 0 \
				-variable ::plugins::iepuos::vars::puo_$puon \
				-command  "::plugins::iepuos::toggle_puo $puon" \
				-pady 0 -bd $::config(min_bd) -anchor w
		if {[expr ($puon/5) % 2]} {
			$w.f1.ft.puo.puo$puon configure -background gray80
		}
		grid $w.f1.ft.puo.puo$puon -column 0 -row [expr 10+$puon] -sticky ew
		incr puon
	}

	frame $w.f1.ft.fc -bd 2 -relief sunken
	pack  $w.f1.ft.fc -side left -fill both -expand 1
	canvas $w.f1.ft.fc.c -xscrollcommand "$w.f1.hs set" -height 8 \
			-highlightcolor [$w.f1.ft.fc cget -background]
	pack   $w.f1.ft.fc.c -expand 1 -fill both
	frame  $w.f1.ft.fc.c.f
	pack   $w.f1.ft.fc.c.f
	scrollbar $w.f1.hs -command "$w.f1.ft.fc.c xview" -orient h
	pack   $w.f1.hs -fill x

	$w.f1.ft.fc.c create window 0 0 -window $w.f1.ft.fc.c.f -anchor nw

	set col 0
	foreach i $vcids_list {
		set vid [lindex $i 0]
		set cid [lindex $i 1]
		set ::plugins::iepuos::vars::edited($vid,$cid) 0

		label $w.f1.ft.fc.c.f.lab_$col -text "$vid/$cid" -pady 0
		if {$::config(use_preview)} {
			button $w.f1.ft.fc.c.f.pre_$col -image ::preview::bitmapplay -width 13 -height 15 \
				-command "::preview::preview_vcid $type $vts $vid $cid false [lindex [::streams::get_current_best_audio_stream] 0]"
		}
		checkbutton $w.f1.ft.fc.c.f.sel_${vid}_$cid \
				-variable ::plugins::iepuos::vars::sel_${vid}_${cid} \
				-pady 0 -bd $::config(min_bd)
		set ::plugins::iepuos::vars::sel_${vid}_${cid} 1
		grid  $w.f1.ft.fc.c.f.lab_$col -column $col -row 1 -sticky ew
		grid  $w.f1.ft.fc.c.f.pre_$col -column $col -row 2 -sticky ew
		grid  $w.f1.ft.fc.c.f.sel_${vid}_$cid -column $col -row 3 -sticky ew
		if {[expr ($col/5) % 2]} {
			$w.f1.ft.fc.c.f.lab_$col configure -background gray80
			catch {$w.f1.ft.fc.c.f.pre_$col configure -background gray80}
			$w.f1.ft.fc.c.f.sel_${vid}_$cid configure -background gray80
		}

		for {set puon 0} {$puon < 25} {incr puon} {
			checkbutton $w.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon -pady 0 \
					-variable ::plugins::iepuos::vars::puo_${vid}_${cid}_$puon \
					-tristatevalue -1 -foreground lightBlue \
					-command "::plugins::iepuos::no_tristate $vid $cid $puon" \
					-pady 0 -bd $::config(min_bd)
			if {[expr ($puon/5) % 2]} {
				if {[expr ($col/5) % 2]} {
					$w.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -background gray70
				} else {
					$w.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -background gray80
				}
			} elseif {[expr ($col/5) % 2]} {
				$w.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -background gray80
			}
			set ::plugins::iepuos::vars::puo_${vid}_${cid}_$puon -1
			grid  $w.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon -column $col -row [expr 10+$puon] -sticky ew
		}
		grid columnconfigure $w.f1.ft.fc.c.f $col -uniform 1
		incr col
	}
	set ::plugins::iepuos::vars::numvcids $col
	frame $w.f1.ft.fc.c.f.sep -width 10 -height 2 -bd 2 -relief sunken
	grid  $w.f1.ft.fc.c.f.sep -column 0 -columnspan $col -row 4 -sticky ew

	### buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 2 -fill x

	menubutton $w.fb.mnu -menu $w.fb.mnu.setvcidpuosmenu_select -text "Select" -pady 0
	pack   $w.fb.mnu -side left -padx 1 -fill y
	menu $w.fb.mnu.setvcidpuosmenu_select
	$w.fb.mnu.setvcidpuosmenu_select add command -label "Select all cells" -command "set ::plugins::iepuos::vars::all_vcids 1; ::plugins::iepuos::select_all_vcids"
	$w.fb.mnu.setvcidpuosmenu_select add command -label "Deselect all cells" -command "set ::plugins::iepuos::vars::all_vcids 0; ::plugins::iepuos::select_all_vcids"
	$w.fb.mnu.setvcidpuosmenu_select add command -label "Toggle selection" -command "::plugins::iepuos::toggle_selection"
	$w.fb.mnu.setvcidpuosmenu_select add separator
	$w.fb.mnu.setvcidpuosmenu_select add command -label "Select only cells with set PUOs" -command "::plugins::iepuos::select_set_cells"
	$w.fb.mnu.setvcidpuosmenu_select add command -label "Select only edited cells" -command "::plugins::iepuos::select_edited_cells"
	$w.fb.mnu.setvcidpuosmenu_select add separator
	set state normal
	if {! $::plugins::iepuos::vars::angle(all)} {set state disabled}
	$w.fb.mnu.setvcidpuosmenu_select add command -state $state -label "Select only multi-angle cells"  -command "set ::plugins::iepuos::vars::all_vcids 0; ::plugins::iepuos::select_all_vcids; ::plugins::iepuos::select_multiangle_cells 1"
	$w.fb.mnu.setvcidpuosmenu_select add command -state $state -label "Add multi-angle cells" -command "::plugins::iepuos::select_multiangle_cells 1"
	$w.fb.mnu.setvcidpuosmenu_select add command -state $state -label "Remove multi-angle cells" -command "::plugins::iepuos::select_multiangle_cells 0"
	$w.fb.mnu.setvcidpuosmenu_select add separator
	set state disabled
	foreach i [array names ::buttons::butsinfo $type,$vts,*,numbuts] {
		if {$::buttons::butsinfo($i) > 0} {set state normal; break}
	}
	$w.fb.mnu.setvcidpuosmenu_select add command -state $state -label "Select only cells with buttons" -command "set ::plugins::iepuos::vars::all_vcids 0; ::plugins::iepuos::select_all_vcids; ::plugins::iepuos::select_button_cells $type $vts 1"
	$w.fb.mnu.setvcidpuosmenu_select add command -state $state -label "Add cells with buttons" -command "::plugins::iepuos::select_button_cells $type $vts 1"
	$w.fb.mnu.setvcidpuosmenu_select add command -state $state -label "Remove cells with buttons" -command "::plugins::iepuos::select_button_cells $type $vts 0"

	menubutton $w.fb.edit -menu $w.fb.edit.setvcidpuosmenu_edit -text "Edit" -pady 0
	pack   $w.fb.edit -side left -padx 1 -fill y
	menu $w.fb.edit.setvcidpuosmenu_edit
	$w.fb.edit.setvcidpuosmenu_edit add command -label "Clear all PUOs of selected cells"  -command "::plugins::iepuos::set_puos 0"
	$w.fb.edit.setvcidpuosmenu_edit add command -label "Set all PUOs of selected cells"    -command "::plugins::iepuos::set_puos 1"
	$w.fb.edit.setvcidpuosmenu_edit add command -label "Toggle all PUOs of selected cells" -command "::plugins::iepuos::set_puos -1"
	$w.fb.edit.setvcidpuosmenu_edit add separator
	$w.fb.edit.setvcidpuosmenu_edit add command -label "Copy PUOs of first selected cell" -command "::plugins::iepuos::copy_puos"
	$w.fb.edit.setvcidpuosmenu_edit add command -label "Paste PUOs to selected cells" -command "::plugins::iepuos::paste_puos"
	$w.fb.edit.setvcidpuosmenu_edit add separator
	$w.fb.edit.setvcidpuosmenu_edit add command -label "Input PUOs value for selected cells" -command "::plugins::iepuos::input_puos_value"

	pack [frame $w.fb.pad] -side left -fill x -padx 1 -expand 1

	button $w.fb.get   -text " Get PUOs from VOBs " -command "::plugins::iepuos::get_from_vobs $type $vts"
	pack   $w.fb.get   -side left -pady 1 -padx 2

	button $w.fb.write -text " Write PUOs to VOBs " -command "::plugins::iepuos::write_to_vobs $type $vts"
	pack   $w.fb.write -side left -padx 1 -padx 2

	button $w.fb.cancel -text "Close" -width 10 -command {set ::plugins::iepuos::vars::done 0}
	pack   $w.fb.cancel -side left -padx 2
	if {$::config(platform) == "mac"} {
		pack [frame $w.fb.macpad] -side left -padx 6
	}

	wm title $w "VOB PUO Editor"

	# inform scrollbar of the width of the canvas's internal window
	update
	$w.f1.ft.fc.c configure -width 30 -height [winfo height $w.f1.ft.fc.c.f]
	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]

	# inform scrollbar of the width of the canvas's internal window
	$w.f1.ft.fc.c configure -width [winfo width $w.f1.ft.fc.c.f] \
			-scrollregion "0 0 [winfo width $w.f1.ft.fc.c.f] 0"

	# window geometry
	update
	set ww [winfo reqwidth $w]
	set max [expr [winfo screenwidth $w]-40]
	if {$ww > $max} {
		$w.f1.ft.fc.c configure -width [expr [winfo width $w.f1.ft.fc.c.f]-($ww-$max)]
		update
		set ww [winfo reqwidth $w]
	}
	update
	set wh [expr [winfo reqheight $w]+4]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-$ww)/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-$wh)/2)-25}]
	if {[expr $wx + $ww + 10] > [winfo screenwidth $w]} {set wx [expr [winfo screenwidth $w]-$ww-10]}
	if {$wx < 10} {set wx 10}
	if {$wy < 30} {set wy 30}
	wm geometry $w ${ww}x$wh+$wx+$wy
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm resizable $w 1 0

	# Done building window
	update idletasks
	wm deiconify $w

	set ::plugins::iepuos::vars::done -1
	tkwait variable ::plugins::iepuos::vars::done

	# verify if there are still cells to save
	select_edited_cells
	set numcells [how_many_selected_cells false]
	if {$numcells > 0} {
		set msg "$numcells cells have been edited but not saved.\nI have selected the edited cells for you.\n\nDo you want to save them before quitting?"
		set r [tk_messageBox -message $msg -window .setvcidpuosnocellwarning \
				-parent .setvcidpuos -icon question -type yesno]
		if {$r == "yes"} {
			write_to_vobs $type $vts
		}
	}

	destroy $w
	destroy $w.fb.mnu.setvcidpuosmenu_select
	destroy $w.fb.edit.setvcidpuosmenu_edit
	namespace delete ::plugins::iepuos::vars
	return ""
}

proc no_tristate {vid cid puon} {
	set ::plugins::iepuos::vars::edited($vid,$cid) 1
	.setvcidpuos.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -foreground black
}

proc toggle_puo {puon} {
	if {[how_many_selected_cells true] == 0} return

	set w .setvcidpuos
	for {set c 0} {$c < $::plugins::iepuos::vars::numvcids} {incr c} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$c cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		if {[subst \$::plugins::iepuos::vars::sel_${vid}_${cid}]} {
			set ::plugins::iepuos::vars::edited($vid,$cid) 1
			set val [subst \$::plugins::iepuos::vars::puo_$puon]
			set ::plugins::iepuos::vars::puo_${vid}_${cid}_$puon $val
			.setvcidpuos.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -foreground black
		}
	}
}

proc copy_puos {} {
	if {[how_many_selected_cells true "You must select a cell to get the PUOs from!"] == 0} return

	set w .setvcidpuos
	set copied false
	set puos {}
	for {set c 0} {$c < $::plugins::iepuos::vars::numvcids} {incr c} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$c cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		if {[subst \$::plugins::iepuos::vars::sel_${vid}_${cid}]} {
			for {set puon 0} {$puon <= 24} {incr puon} {
				set val [subst \$::plugins::iepuos::vars::puo_${vid}_${cid}_$puon]
				lappend puos $val
				if {$val != -1} {
					set copied true
				}
			}
			break
		}
	}
	if {$copied} {
		lappend puos "PUOs"
		clipboard clear
		clipboard append $puos
	} else {
		bell
	}
}

proc paste_puos {} {
	if {[how_many_selected_cells true] == 0} return

	set puos [clipboard get]
	if {[llength $puos] != 26 || [lindex $puos 25] != "PUOs"} {
		bell
		return
	}
	set w .setvcidpuos
	for {set c 0} {$c < $::plugins::iepuos::vars::numvcids} {incr c} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$c cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		if {[subst \$::plugins::iepuos::vars::sel_${vid}_${cid}]} {
			for {set puon 0} {$puon <= 24} {incr puon} {
				set val [lindex $puos $puon]
				if {$val != -1} {
					set ::plugins::iepuos::vars::puo_${vid}_${cid}_$puon $val
					.setvcidpuos.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -foreground black
				}
			}
		}
	}
	for {set puon 0} {$puon <= 24} {incr puon} {
		set val [lindex $puos $puon]
		if {$val != -1} {
			set ::plugins::iepuos::vars::puo_$puon $val
		}
	}
}

proc input_puos_value {} {
	if {[how_many_selected_cells true] == 0} return

	set w .setvcidpuos
	set puosval [::stringreq::stringreq -window .setvcidpuos_inputpuos -parent $w \
			-title "PUOs value" -def 0 -width 10 -geometry c \
			-callback ::pgced::validate_input_puo_value]
	grab set $w
	if {$puosval == ""} return
	for {set puon 0} {$puon <= 24} {incr puon} {
		if {[expr {$puosval & round(pow(2,$puon))}] > 0} {
			set val 1
		} else {
			set val 0
		}
		set ::plugins::iepuos::vars::puo_$puon $val
		set puo($puon) $val
	}
	for {set c 0} {$c < $::plugins::iepuos::vars::numvcids} {incr c} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$c cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		if {[subst \$::plugins::iepuos::vars::sel_${vid}_${cid}]} {
			for {set puon 0} {$puon <= 24} {incr puon} {
				set ::plugins::iepuos::vars::puo_${vid}_${cid}_$puon $puo($puon)
				.setvcidpuos.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -foreground black
			}
		}
	}
}

# state can be 0 or 1 to unset or set the PUOs, or -1 to toggle it.
# Undefined PUOs are ignored when state == -1
proc set_puos {state} {
	if {[how_many_selected_cells true] == 0} return

	set w .setvcidpuos
	for {set c 0} {$c < $::plugins::iepuos::vars::numvcids} {incr c} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$c cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		if {[subst \$::plugins::iepuos::vars::sel_${vid}_${cid}]} {
			set ::plugins::iepuos::vars::edited($vid,$cid) 1
			for {set puon 0} {$puon <= 24} {incr puon} {
				if {$state == -1} {
					set curstate [subst \$::plugins::iepuos::vars::puo_${vid}_${cid}_$puon]
					if {$curstate != -1} {
						set ::plugins::iepuos::vars::puo_${vid}_${cid}_$puon [expr 1 - $curstate]
						.setvcidpuos.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -foreground black
					}
				} else {
					set ::plugins::iepuos::vars::puo_${vid}_${cid}_$puon $state
					.setvcidpuos.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -foreground black
				}
			}
		}
	}
	if {$state != -1} {
		for {set puon 0} {$puon <= 24} {incr puon} {
			set ::plugins::iepuos::vars::puo_$puon $state
		}
	}
}

proc select_all_vcids {} {
	set w .setvcidpuos
	for {set c 0} {$c < $::plugins::iepuos::vars::numvcids} {incr c} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$c cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		set ::plugins::iepuos::vars::sel_${vid}_${cid} $::plugins::iepuos::vars::all_vcids
	}
}

# scope can be "all" or "selected"
proc write_to_vobs {type vts} {
	set w .setvcidpuos

	if {[how_many_selected_cells false] == 0} {
		select_edited_cells
		set numcells [how_many_selected_cells false]
		if {$numcells == 0} {
			set msg "No cells are selected, and no cells have been edited.\n\nThere is nothing to save."
			tk_messageBox -message $msg -window .setvcidpuosnocellwarning \
				-parent .setvcidpuos -icon warning
			return
		} else {
			set msg "No cells were selected, but $numcells cells have been edited.\nI have selected the edited cells for you.\n\nDo you want to save them?"
			set r [tk_messageBox -message $msg -window .setvcidpuosnocellwarning \
					-parent .setvcidpuos -icon question -type yesno]
			if {$r == "no"} return
		}
	}
	set numcells 0
	for {set c 0} {$c < $::plugins::iepuos::vars::numvcids} {incr c} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$c cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		if {[subst \$::plugins::iepuos::vars::sel_${vid}_${cid}]} {
			incr numcells $::plugins::iepuos::vars::edited($vid,$cid)
		}
	}
	if {$numcells == 0} {
		set msg "None of the selected cells have been edited.\n\nThere is nothing to save."
		tk_messageBox -message $msg -window .setvcidpuosnocellwarning \
			-parent .setvcidpuos -icon warning
		return
	}

	set progress 0
	set old_vobfile "-"
	set fh ""
	set global_error false
	if {$type == "menu"} {
		if {$vts == 0} {
			set domain "VMGM"
		} else {
			set domain "VTSM $vts"
		}
	} else {
		set domain "VTST $vts"
	}
	set msg "Saving edited PUOs of the selected cells of the $domain."
	show_msg $msg
	set blocking [blocking_msg $msg true]

	set aborted false
	for {set cell 0} {$cell < $::plugins::iepuos::vars::numvcids} {incr cell} {
		if {$aborted || $::blocking_msg_aborted} break
		incr progress
		set vcid [split [$w.f1.ft.fc.c.f.lab_$cell cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		if {! [subst \$::plugins::iepuos::vars::sel_${vid}_${cid}]} {
			continue
		}
		if {! $::plugins::iepuos::vars::edited($vid,$cid)} {
			show_msg "Skipping non-edited cell $vid/$cid." $progress $::plugins::iepuos::vars::numvcids false false
			continue
		}

		set val_or  0
		set val_and [expr 0xFFFFFFFF]
		set saved_puos -1
		for {set puon 0} {$puon <= 24} {incr puon} {
			set val [expr round(pow(2,$puon))]
			set flag [subst \$::plugins::iepuos::vars::puo_${vid}_${cid}_$puon]
			if {$flag == 1} {
				set val_or [expr $val_or + $val]
			} elseif {$flag == 0} {
				set val_and [expr $val_and - $val]
			}
		}

		show_msg "Writing PUOs of cell $vid/$cid..." $progress $::plugins::iepuos::vars::numvcids false false
		set lba [lindex $::plugins::iepuos::vars::start_end($vid,$cid) 0]
		set last_lba [lindex $::plugins::iepuos::vars::start_end($vid,$cid) 1]
		set err false
		set done false
		set check_for_abort 0
		while {$lba <= $last_lba && ! $done && ! $err} {
			incr check_for_abort
			if {$check_for_abort == 20 && ! $aborted} {
				set check_for_abort 0
				after 1 {set ::plugins::iepuos::vars::pause 1}
				tkwait variable ::plugins::iepuos::vars::pause
				if {$::blocking_msg_aborted && ! $aborted} {
					update idletasks
					after 200 {set ::plugins::iepuos::vars::pause 1}
					tkwait variable ::plugins::iepuos::vars::pause
					close_blocking_msg 1
					grab set $w
					blocking_msg "Abort request received!\nFinishing current cell..."
					set aborted true
				}
			}
			set poslist [::menued::get_nav_pack_position $type $vts $lba $::dvddir]
			set vobfile [lindex $poslist 0]
			set sector  [lindex $poslist 1]
			if {$vobfile == ""} {
				set err true
				break
			}
			if {$vobfile != $old_vobfile} {
				if {![file exists $vobfile]} {
					set err true
					break
				}
				if {$fh != ""} {
					close $fh
					set old_vobfile ""
					set ::buttons::butsinfo($type,$vts,[file tail $vobfile],mtime) [file mtime $vobfile]
				}
				set fh [::menued::openbinfile $vobfile r+]
				if {$fh == ""} {
					set err true
					break
				}
				set old_vobfile $vobfile
			}
			set navpack [::menued::load_nav_pack $vobfile $fh $sector]
			if {$navpack == ""} {
				set err true
				break
			}
			set v [get_int navpack 0x41F 2]
			set c [get_int navpack 0x422 1]
			if {$v == $vid && $c == $cid} {
				set puos [get_int navpack 0x35 4]
				set newpuos [expr ($puos & $val_and) | $val_or]
				set_int navpack 0x35 $newpuos 4
				if {! [::menued::save_nav_pack $vobfile $fh $sector $navpack]} {
					set err true
					break
				}
				if {$lba == $::plugins::iepuos::vars::lbas($vid,$cid)} {
					set saved_puos $newpuos
				}
			}
			set lbaoffset [get_int navpack 1039 4]
			if {$lbaoffset <= 0} {set done true}
			set lba [expr {$lba+$lbaoffset+1}]
			if {$lba <= 0} {set done true}
		}
		if {$err} {
			for {set puon 0} {$puon <= 24} {incr puon} {
				.setvcidpuos.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -background pink
			}
			show_msg "Error loading or writing nav pack of cell $vid/$cid at LBA $lba (sector $sector in file [file tail $vobfile])!"
			set global_error true
		} elseif {$saved_puos == -1} {
			for {set puon 0} {$puon <= 24} {incr puon} {
				.setvcidpuos.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -background orange
			}
			show_msg "Unknown error when saving the PUOs of cell $vid/$cid.  Wrong VOB/Cell ID?"
			set global_error true
		} else {
			for {set puon 0} {$puon <= 24} {incr puon} {
				.setvcidpuos.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -foreground black
				if {[expr {$saved_puos & round(pow(2,$puon))}] > 0} {
					set ::plugins::iepuos::vars::puo_${vid}_${cid}_$puon 1
				} else {
					set ::plugins::iepuos::vars::puo_${vid}_${cid}_$puon 0
				}
			}
			set ::plugins::iepuos::vars::edited($vid,$cid) 0
		}
		after 50 {set ::plugins::iepuos::vars::pause 1}
		tkwait variable ::plugins::iepuos::vars::pause
	}
	if {$fh != ""} {
		close $fh
		set ::buttons::butsinfo($type,$vts,[file tail $vobfile],mtime) [file mtime $vobfile]
	}

	if {[catch {close_blocking_msg $blocking}]} {close_blocking_msg 1}
	grab set $w

	if {$global_error} {
		show_msg "Errors when writing PUOs in VOBs of $domain!" 0 100 true true
		tk_messageBox -message "Some errors occurred when loading or saving nav-packs.\nThe concerned cells have been highlighted in red.\nSee the PgcEdit log for details." \
				-icon error -title "VOB PUO Editor" -parent $w -window ${w}err
		return false
	} else {
		show_msg "Done writing PUOs in VOBs of $domain." 0 100 true false
		# verify if the user has edited the menu buttons or BOVs
		set edited false
		foreach i [array names ::buttons::butsinfo *,edited] {
			if {$::buttons::butsinfo($i)} {
				set edited true
				break
			}
		}
		# and save the button info file if no buttons have been modified
		if {[info exists ::backupdir] && \
				 (! [info exists ::menubuttons::buttons_loaded_from_button_file] || \
				 $::menubuttons::buttons_loaded_from_button_file == "")} {
			if {[file exists [file join $::backupdir "menubuttons_new.but"]]} {
				::menubuttons::save_buts [file join $::backupdir "menubuttons_new.but"]
			} elseif {[file exists [file join $::backupdir "menubuttons.but"]]} {
				::menubuttons::save_buts [file join $::backupdir "menubuttons.but"]
			}
		}

		return true
	}
}

proc get_from_vobs {type vts} {
	if {[how_many_selected_cells true] == 0} return

	set w .setvcidpuos

	set numcells 0
	for {set cell 0} {$cell < $::plugins::iepuos::vars::numvcids} {incr cell} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$cell cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		if {[subst \$::plugins::iepuos::vars::sel_${vid}_${cid}]} {
			incr numcells $::plugins::iepuos::vars::edited($vid,$cid)
		}
	}
	if {$numcells > 0} {
		set msg "$numcells currently selected cells have been edited but not saved.\nIf you continue, your edits will be lost!\n(Use Select -> Select Only Edited Cells to see which cells have been edited.)\n\nRead the PUOs values of the selected cells from the VOBs anyway?"
		set r [tk_messageBox -message $msg -window .setvcidpuosnocellwarning \
				-parent .setvcidpuos -icon question -type yesno -default no]
		if {$r == "no"} return
	}

	if {$type == "menu"} {
		if {$vts == 0} {
			set domain "VMGM"
		} else {
			set domain "VTSM $vts"
		}
	} else {
		set domain "VTST $vts"
	}
	set blocking [blocking_msg "Getting PUOs from VOBs of the $domain."]
	set old_vobfile "-"
	set fh ""
	for {set cell 0} {$cell < $::plugins::iepuos::vars::numvcids} {incr cell} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$cell cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		if {! [subst \$::plugins::iepuos::vars::sel_${vid}_${cid}]} continue

		set lba $::plugins::iepuos::vars::lbas($vid,$cid)

		set err false
		set poslist [::menued::get_nav_pack_position $type $vts $lba $::dvddir]
		set vobfile [lindex $poslist 0]
		set sector  [lindex $poslist 1]
		if {$vobfile == ""} {
			set err true
		}
		if {! $err && $vobfile != $old_vobfile} {
			if {![file exists $vobfile]} {
				set err true
			}
			if {$fh != ""} {
				close $fh
				set old_vobfile ""
			}
			set fh [::menued::openbinfile $vobfile r]
			if {$fh == ""} {
				set err true
			}
			set old_vobfile $vobfile
		}
		if {! $err} {
			set navpack [::menued::load_nav_pack $vobfile $fh $sector]
			if {$navpack == ""} {
				set err true
			}
		}
		if {$err} {
			for {set puon 0} {$puon <= 24} {incr puon} {
				.setvcidpuos.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -foreground lightBlue
				set ::plugins::iepuos::vars::puo_${vid}_${cid}_$puon -1
			}
		} else {
			set puos [get_int navpack 0x35 4]
			set v [get_int navpack 0x41F 2]
			set c [get_int navpack 0x422 1]
			if {$v == $vid && $c == $cid} {
				for {set puon 0} {$puon <= 24} {incr puon} {
					.setvcidpuos.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -foreground black
					if {[expr {$puos & round(pow(2,$puon))}] > 0} {
						set ::plugins::iepuos::vars::puo_${vid}_${cid}_$puon 1
					} else {
						set ::plugins::iepuos::vars::puo_${vid}_${cid}_$puon 0
					}
				}
				set ::plugins::iepuos::vars::edited($vid,$cid) 0
			} else {
				for {set puon 0} {$puon <= 24} {incr puon} {
					.setvcidpuos.f1.ft.fc.c.f.puo_${vid}_${cid}_$puon configure -background pink
				}
				show_msg "Wrong VOB/Cell ID for cell $vid/$cid at LBA $lba ($sector in file [file tail $vobfile]: V/CID in navpack is $v/$c!" -1 -1 true true
			}
		}
		update idletasks
		after 10 {set ::pause 1}
		tkwait variable ::pause
	}
	if {$fh != ""} {
		close $fh
	}
	if {[catch {close_blocking_msg $blocking}]} {close_blocking_msg 1}
	grab set $w
}


proc select_multiangle_cells {state} {
	set w .setvcidpuos
	for {set c 0} {$c < $::plugins::iepuos::vars::numvcids} {incr c} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$c cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		if {$::plugins::iepuos::vars::angle($vid,$cid)} {
			set ::plugins::iepuos::vars::sel_${vid}_${cid} $state
		}
	}
}

proc select_button_cells {type vts state} {
	set w .setvcidpuos
	for {set c 0} {$c < $::plugins::iepuos::vars::numvcids} {incr c} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$c cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		if {[info exists ::buttons::butsinfo($type,$vts,$vid,$cid,numbuts)] && \
				$::buttons::butsinfo($type,$vts,$vid,$cid,numbuts) > 0}  {
			set ::plugins::iepuos::vars::sel_${vid}_${cid} $state
		}
	}
}

proc toggle_selection {} {
	set w .setvcidpuos
	for {set c 0} {$c < $::plugins::iepuos::vars::numvcids} {incr c} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$c cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		set ::plugins::iepuos::vars::sel_${vid}_${cid} \
				[expr 1 - [subst \$::plugins::iepuos::vars::sel_${vid}_${cid}]]
	}
}

proc select_set_cells {} {
	set w .setvcidpuos
	for {set c 0} {$c < $::plugins::iepuos::vars::numvcids} {incr c} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$c cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		set ::plugins::iepuos::vars::sel_${vid}_${cid} 0
		for {set puon 0} {$puon <= 24} {incr puon} {
			if {[subst \$::plugins::iepuos::vars::puo_${vid}_${cid}_$puon] == 1} {
				set ::plugins::iepuos::vars::sel_${vid}_${cid} 1
				break
			}
		}
	}
}

proc select_edited_cells {} {
	set w .setvcidpuos
	for {set c 0} {$c < $::plugins::iepuos::vars::numvcids} {incr c} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$c cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		set ::plugins::iepuos::vars::sel_${vid}_${cid} $::plugins::iepuos::vars::edited($vid,$cid)
	}
}

proc how_many_selected_cells {{warn_if_empty false} {msg "No cells are selected!"}} {
	set w .setvcidpuos
	set num 0
	for {set c 0} {$c < $::plugins::iepuos::vars::numvcids} {incr c} {
		set vcid [split [$w.f1.ft.fc.c.f.lab_$c cget -text] /]
		set vid [lindex $vcid 0]
		set cid [lindex $vcid 1]
		incr num [subst \$::plugins::iepuos::vars::sel_${vid}_${cid}]
	}
	if {$warn_if_empty && $num == 0} {
		tk_messageBox -message $msg -window .setvcidpuosnocellwarning \
				-parent .setvcidpuos -icon warning
	}
	return $num
}


}	;# end of namespace ::plugins::iepuos
